"""Mixins for Redis connectivity."""


from typing import Dict
from redis import Redis


class RedisMixin:
    """Mixin for giving Redis connectivity to an object."""

    _borg_url: Dict[str, str] = {"url": ""}
    _borg_connections: Dict[str, Redis] = {}

    def connect(self, url: str):
        """Set the redis URL for the borg."""
        self._borg_url["url"] = url

    @property
    def redis(self) -> Redis:
        """Get the Redis instance from the borg."""
        url = self._borg_url["url"]
        assert (
            url
        ), "Please call `.connect` on any instance to initialize the connection."
        client = self._borg_connections.get(url)

        if client is None:
            created = Redis.from_url(url)
            self._borg_connections[url] = created
            return created

        return client



"""The event class.

Events are JSON-compatible objects stored in Redis streams, to be consumed
by applications.

Any JSON object can be turned into an Event.
"""


from typing import Union, NamedTuple, Dict, List, Any
import json


class Event(NamedTuple):
    """An event sent through a Redis stream."""

    data: Union[Dict[str, Any], List[Any], None]
    ruid: Union[str, None] = None

    def __str__(self):
        return f"<Event ruid={str(self.ruid)}>"

    @classmethod
    def deserialize(
        cls, raw: Dict[bytes, bytes], ruid: Union[bytes, None] = None
    ) -> "Event":
        """Deserialize a raw dictionary to an event."""
        data = json.loads(raw[b"topix:event.data"].decode("utf-8"))
        if ruid:
            return cls(data, ruid=ruid.decode("utf-8"))
        return cls(data)

    def serialize(self) -> Dict[str, bytes]:
        """Serialize to a Redis Streams-compatible dictionary."""
        return {"topix:event.data": json.dumps(self.data).encode("utf-8")}



"""A Topic is a specific Redis Stream that holds Events."""


from typing import Union, Dict, List, Any
import logging

import redis

from topix.mixins import RedisMixin
from topix.event import Event


LOG = logging.getLogger(__name__)


class Topic(RedisMixin):
    """A Topic is analagous to a Redis `stream` with an easy to use Python interface.

    Usage:
    >>> topic = Topic("incoming_webhooks")
    >>> topic.add({"user": 18, "action": "create_order"})

    Often times, you'll want to use a Topic alongside a consumer class.
    """

    def __init__(self, name: str):
        self.name = name

    def add(self, event: Union[Dict[str, Any], List[Any], Event]) -> bytes:
        """Add an event to the event stream."""
        # Duck type check for event classes.
        try:
            data = event.serialize()  # type: ignore
        except AttributeError:
            data = Event(event).serialize()  # type: ignore
        return self.redis.xadd(self.name, data)

    def create_group(self, group: str):
        """Idempotently create a consumer group.

        If the group already exists, simply pass.
        """
        try:
            self.redis.xgroup_create(self.name, group, mkstream=True)
        except redis.exceptions.ResponseError:
            LOG.debug(f"Consumer group {group} for {self.name} already exists")

    def read(
        self, group: str, consumer: str, timeout=0, seen=">"
    ) -> Union[List[Event], None]:
        """Read latest events from the stream for a given consumer group.

        Arguments:
            group: The name of the group.
            consumer: Consumer ID as a string.
        Keyword Arguments:
            timeout: The timeout in milliseconds, or 0 for "block"
            seen: The last ID seen OR the string ">" for "latest".
        """
        LOG.debug(f"Reading latest messages from {self.name} for {group}.{consumer}")
        data = self.redis.xreadgroup(group, consumer, {self.name: seen}, block=timeout)
        # If it's blank or empty, let the caller wait again with a clear signal.
        if not data:
            return None

        # We already know the stream. Discard it.
        _, keyvals = data[0]
        return [Event.deserialize(event, ruid) for (ruid, event) in keyvals]

    def ack(self, event: Event, consumer: str):
        """Acknowledge the success of a given event.

        Arguments:
            event: An Event.
            consumer: ID of the acking consumer.
        """
        assert event.ruid, "Event has no .ruid, was it deserialized?"
        LOG.debug(f"Consumer {consumer} acks {event}")
        self.redis.xack(self.name, consumer, event.ruid)



"""Consumer class for working on streams."""


import uuid
from typing import Union, List, Generator
from functools import wraps

from topix.topic import Topic
from topix.event import Event


def requires_setup(func):
    """Wraps methods that require setup."""

    @wraps(func)
    def wrapped(self, *args, **kwargs):
        if not self.is_setup:
            self.setup()
        return func(self, *args, **kwargs)

    return wrapped


class Consumer:
    """A consumer is a registered consumer of a stream, belonging to a group."""

    def __init__(
        self,
        stream: str,
        group: str,
        guid: str = None,
        topic: Union[Topic, None] = None,
    ):
        self.stream_name = stream
        self.group_name = group
        self.guid = guid if guid else str(uuid.uuid4())

        self.topic = topic if topic else Topic(self.stream_name)
        self.is_setup = False

    def setup(self):
        """Run setup methods for the consumer."""
        self.topic.create_group(self.group_name)
        self.is_setup = True

    @requires_setup
    def ack(self, event: Event):
        """Ack an event."""
        self.topic.ack(event, self.guid)

    @requires_setup
    def read(
        self, timeout: int = 0, seen: Union[str, int] = ">"
    ) -> Union[List[Event], None]:
        """Listen for new entries in the stream"""
        return self.topic.read(self.group_name, self.guid, timeout=timeout, seen=seen)

    @requires_setup
    def yield_from_stream(
        self, many=True
    ) -> Generator[Union[List[Event], Event], None, None]:
        """Yield groups of events (or single events) from the stream.

        Keyword Arguments:
            many: Whether to yield lists of events or single events. (default True)
        """
        events = self.read()
        if many:
            yield events
        else:
            for event in events:
                yield event
